# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_core.proto import klio_pb2

from klio.transforms import decorators
from klio.utils import _thread_limiter


@pytest.fixture
def kmsg():
    msg = klio_pb2.KlioMessage()
    msg.data.element = b"3l3m3nt"
    return msg


def test_retry(kmsg, mocker, mock_config):
    mock_function = mocker.Mock()

    @decorators._handle_klio
    @decorators._retry(tries=2)
    def func(*args, **kwargs):
        mock_function(*args, **kwargs)
        raise Exception("fuu")

    func(kmsg.SerializeToString())

    assert 2 == mock_function.call_count


def test_retry_custom_catch(kmsg, mocker, mock_config):

    # Assert retry on custom exception
    class CustomCatchException(Exception):
        pass

    mock_function = mocker.Mock()
    raise_custom = True

    @decorators._handle_klio
    @decorators._retry(tries=3, exception=CustomCatchException)
    def func(*args, **kwargs):
        mock_function(*args, **kwargs)
        if raise_custom:
            raise CustomCatchException("custom fuu")
        raise Exception("fuu")

    func(kmsg.SerializeToString())
    assert 3 == mock_function.call_count

    # Assert retry on multiple custom exceptions
    class AnotherCustomCatchException(Exception):
        pass

    exc_tuple = (CustomCatchException, AnotherCustomCatchException)

    @decorators._handle_klio
    @decorators._retry(tries=3, exception=exc_tuple)
    def func(*args, **kwargs):
        mock_function(*args, **kwargs)
        if raise_custom:
            raise CustomCatchException("custom fuu")
        raise Exception("fuu")

    mock_function.reset_mock()
    func(kmsg.SerializeToString())
    assert 3 == mock_function.call_count

    # Assert retries do not happen for an exception that isn't provided
    mock_function.reset_mock()
    raise_custom = False
    func(kmsg.SerializeToString())
    assert 1 == mock_function.call_count


def test_retry_raises_runtime_parents(kmsg, mocker, mock_config):
    # Need to call @retry with parens
    with pytest.raises(RuntimeError):

        @decorators._handle_klio
        @decorators.retry
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())


@pytest.mark.parametrize(
    "invalid_tries", (-2, 0.5, "1", {"a": "dict"}, ["a", "list"], lambda x: x)
)
def test_retry_raises_runtime_invalid_tries(
    invalid_tries, kmsg, mocker, mock_config
):
    # Assert `tries` as a valid integer
    with pytest.raises(RuntimeError):

        @decorators._handle_klio
        @decorators._retry(tries=invalid_tries)
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())


@pytest.mark.parametrize(
    "invalid_delay", (-2, {"a": "dict"}, ["a", "list"], lambda x: x)
)
def test_retry_raises_runtime_invalid_delay(
    invalid_delay, kmsg, mocker, mock_config
):
    # Assert `delay` as a valid int/float
    with pytest.raises(RuntimeError):

        @decorators._handle_klio
        @decorators._retry(tries=1, delay=invalid_delay)
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())


@pytest.mark.parametrize(
    "max_thread_count,patch_str",
    (
        (None, "threading.BoundedSemaphore"),
        (_thread_limiter.ThreadLimit.DEFAULT, "threading.BoundedSemaphore"),
        (_thread_limiter.ThreadLimit.NONE, "_DummySemaphore"),
    ),
)
def test_thread_limiting(
    max_thread_count, patch_str, kmsg, mock_config, mocker, monkeypatch
):
    mock_function = mocker.Mock()
    mock_semaphore = mocker.Mock()
    monkeypatch.setattr(
        f"klio.utils._thread_limiter.{patch_str}", mock_semaphore
    )

    kwargs = {}
    if max_thread_count is not None:
        kwargs["max_thread_count"] = max_thread_count

    @decorators._handle_klio(**kwargs)
    def func(*args, **kwargs):
        mock_function(*args, **kwargs)
        return

    func(kmsg.SerializeToString())

    assert 1 == mock_function.call_count
    mock_semaphore.return_value.acquire.assert_called_once_with()
    mock_semaphore.return_value.release.assert_called_once_with()


def test_thread_limiting_custom_limiter(
    kmsg, mock_config, mocker, monkeypatch
):
    mock_function = mocker.Mock()
    mock_semaphore = mocker.Mock()

    limiter = _thread_limiter.ThreadLimiter(max_thread_count=1)
    monkeypatch.setattr(limiter, "_semaphore", mock_semaphore)

    @decorators._handle_klio(thread_limiter=limiter)
    def func(*args, **kwargs):
        mock_function(*args, **kwargs)
        return

    func(kmsg.SerializeToString())

    assert 1 == mock_function.call_count
    mock_semaphore.acquire.assert_called_once_with()
    mock_semaphore.release.assert_called_once_with()


def test_thread_limiting_raises_mutex_args(kmsg, mocker, mock_config):
    limiter = _thread_limiter.ThreadLimiter(max_thread_count=1)

    with pytest.raises(RuntimeError):

        @decorators._handle_klio(max_thread_count=1, thread_limiter=limiter)
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())


def test_thread_limiting_raises_invalid_limiter(kmsg, mocker, mock_config):
    limiter = "not an instance of ThreadLimiter"

    with pytest.raises(RuntimeError):

        @decorators._handle_klio(thread_limiter=limiter)
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())


@pytest.mark.parametrize(
    "invalid_max_thread_count", (-2, {"a": "dict"}, ["a", "list"])
)
def test_thread_limiting_raises_invalid_max(
    invalid_max_thread_count, kmsg, mocker, mock_config
):

    with pytest.raises(RuntimeError):

        @decorators._handle_klio(max_thread_count=invalid_max_thread_count)
        def func(*args, **kwargs):
            pass

        func(kmsg.SerializeToString())



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from apache_beam import pvalue
from google.protobuf import message as gproto_message

from klio_core.proto.v1beta1 import klio_pb2

from klio.message import exceptions
from klio.message import serializer


def _get_klio_job():
    job = klio_pb2.KlioJob()
    job.job_name = "klio-job"
    job.gcp_project = "test-project"
    return job


def _get_klio_message():
    parent_klio_job = _get_klio_job()
    msg = klio_pb2.KlioMessage()
    msg.metadata.visited.extend([parent_klio_job])
    msg.metadata.force = True
    msg.metadata.ping = True
    msg.data.element = b"1234567890"
    msg.version = klio_pb2.Version.V2

    return msg


@pytest.fixture
def klio_message():
    return _get_klio_message()


@pytest.fixture
def klio_message_str(klio_message):
    return klio_message.SerializeToString()


@pytest.fixture
def logger(mocker):
    return mocker.Mock()


@pytest.mark.parametrize(
    "version",
    (klio_pb2.Version.UNKNOWN, klio_pb2.Version.V1, klio_pb2.Version.V2),
)
@pytest.mark.parametrize(
    "element,entity_id,payload",
    (
        (b"an-element", None, None),
        (None, "an-entity-id", None),
        (None, "an-entity-id", b"some-payload"),
        (b"an-element", None, b"some-payload"),
        (None, None, b"some-payload"),
    ),
)
def test_handle_msg_compat(version, element, entity_id, payload):
    msg = klio_pb2.KlioMessage()
    msg.version = version
    if element:
        msg.data.element = element
    if payload:
        msg.data.payload = payload
    if entity_id:
        msg.data.entity_id = entity_id

    actual_msg = serializer._handle_msg_compat(msg)
    assert actual_msg.version is not klio_pb2.Version.UNKNOWN
    # we assume in the function's logic that v2 messages are already parsed
    # correctly
    if entity_id and not klio_pb2.Version.V2:
        assert entity_id == actual_msg.data.element.decode("utf-8")


def test_to_klio_message(klio_message, klio_message_str, klio_config, logger):
    actual_message = serializer.to_klio_message(
        klio_message_str, klio_config, logger
    )

    assert klio_message == actual_message
    logger.error.assert_not_called()


def test_to_klio_message_allow_non_kmsg(klio_config, logger, monkeypatch):
    monkeypatch.setattr(
        klio_config.job_config, "allow_non_klio_messages", True
    )
    incoming = b"Not a klio message"
    expected = klio_pb2.KlioMessage()
    expected.data.element = incoming
    expected.version = klio_pb2.Version.V2
    expected.metadata.intended_recipients.anyone.SetInParent()

    actual_message = serializer.to_klio_message(incoming, klio_config, logger)

    assert expected == actual_message
    logger.error.assert_not_called()


def test_to_klio_message_raises(klio_config, logger, monkeypatch):
    incoming = b"Not a klio message"

    with pytest.raises(gproto_message.DecodeError):
        serializer.to_klio_message(incoming, klio_config, logger)

    # Just asserting it's called - not testing the error string itself
    # to avoid making brittle tests
    assert 1 == logger.error.call_count


@pytest.mark.parametrize(
    "payload,exp_payload",
    (
        (None, None),
        (b"some payload", b"some payload"),
        (_get_klio_message().data, None),
        ("string payload", b"string payload"),
    ),
)
def test_from_klio_message(klio_message, payload, exp_payload):
    expected = _get_klio_message()
    if exp_payload:
        expected.data.payload = exp_payload

    expected_str = expected.SerializeToString()

    actual_message = serializer.from_klio_message(klio_message, payload)
    assert expected_str == actual_message


def test_from_klio_message_v1():
    payload = b"some-payload"
    msg = klio_pb2.KlioMessage()
    msg.version = klio_pb2.Version.V1
    msg.data.payload = payload

    expected_str = msg.SerializeToString()

    actual_message = serializer.from_klio_message(msg, payload)
    assert expected_str == actual_message


def test_from_klio_message_tagged_output(klio_message):
    payload = b"some payload"
    expected_msg = _get_klio_message()
    expected_msg.data.payload = payload

    expected = pvalue.TaggedOutput("a-tag", expected_msg.SerializeToString())

    tagged_payload = pvalue.TaggedOutput("a-tag", payload)
    actual_message = serializer.from_klio_message(klio_message, tagged_payload)

    # can't compare expected vs actual directly since pvalue.TaggedOutput
    # hasn't implemented the comparison operators
    assert expected.tag == actual_message.tag
    assert expected.value == actual_message.value


def test_from_klio_message_raises(klio_message):
    payload = {"no": "bytes casting"}

    with pytest.raises(
        exceptions.KlioMessagePayloadException, match="Returned payload"
    ):
        serializer.from_klio_message(klio_message, payload)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import multiprocessing

import pytest

from klio.utils import _thread_limiter


N_CPU = multiprocessing.cpu_count()
BASE_LOG = "Initial semaphore value:"


@pytest.mark.parametrize(
    "max_thread_count, exp_count_log",
    (
        (1, f"{BASE_LOG} 1"),
        (_thread_limiter.ThreadLimit.NONE, "Using unlimited semaphore"),
        (_thread_limiter.ThreadLimit.DEFAULT, f"{BASE_LOG} {N_CPU}"),
        (lambda: 2 * 3, f"{BASE_LOG} 6"),
    ),
)
def test_thread_limiter(max_thread_count, exp_count_log, caplog):
    _thread_limiter.ThreadLimiter(max_thread_count=max_thread_count)

    assert 1 == len(caplog.messages)
    assert exp_count_log in caplog.messages[0]


@pytest.mark.parametrize(
    "max_thread_count", (2, _thread_limiter.ThreadLimit.NONE)
)
def test_thread_limiter_ctx_mgr(max_thread_count, mocker, monkeypatch, caplog):
    mock_semaphore = mocker.Mock()
    limiter = _thread_limiter.ThreadLimiter(max_thread_count=max_thread_count)
    monkeypatch.setattr(limiter, "_semaphore", mock_semaphore)

    with limiter:
        3 * 3

    mock_semaphore.acquire.assert_called_once_with()
    mock_semaphore.release.assert_called_once_with()
    assert 3 == len(caplog.messages)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum
import logging
import multiprocessing
import threading


class _DummySemaphore(object):
    """Mock semaphore for API parity when no limit is set in ThreadLimiter."""

    def __init__(self, *args, **kwargs):
        self._value = "NONE"

    def acquire(self):
        pass

    def release(self):
        pass


class ThreadLimit(enum.Enum):
    """:class:`enum.Enum` of constants for :class:`ThreadLimiter`.

    Example usage:

    .. code-block:: python

        from klio import utils

        thread_limiter = utils.ThreadLimiter(
            max_thread_limit=utils.ThreadLimit.DEFAULT
        )
    """

    NONE = 0
    """Do not limit the number of threads used."""
    DEFAULT = 1
    """Default thread limit (CPU count of worker machine via
    :func:`multiprocessing.cpu_count`).
    """


class ThreadLimiter(object):
    """A context manager to limit active threads for a block of code.

    Threads aren't limited `directly`; threads are limited indirectly
    by maintaining a semaphore (an atomic counter).

    Example usage:

    .. code-block:: python

        # No arg to use the default limit (CPU count of host/worker):
        thread_limiter = ThreadLimiter()

        # Then use it as a context manager:
        with thread_limiter:
            ...

        # Or explicitly call acquire and release:
        thread_limiter.acquire()
        ...
        thread_limiter.release()

        # Explicitly use the default limit:
        thread_limiter = ThreadLimiter(max_thread_count=ThreadLimit.DEFAULT)

        # Set your own limit:
        thread_limiter = ThreadLimiter(max_thread_count=2)

        # Dynamically set the limit using a function
        limit_func = lambda: multiprocessing.cpu_count() * 4
        thread_limiter = ThreadLimiter(max_thread_count=limit_func)

        # Turn off thread limiting
        thread_limiter = ThreadLimiter(max_thread_count=ThreadLimit.NONE)

    Args:
        max_thread_count (int, callable, ThreadLimit): number of threads
            to make available to the limiter, or a :func:`callable` that
            returns an ``int``. Values must be greater or equal to 0.
            Set to :attr:`ThreadLimit.NONE` for no thread limits.
        name (str): Name of particular limiter. Defaults to object ID via
            ``id(self)``.
    """

    _PREFIX = "KlioThreadLimiter"

    def __init__(self, max_thread_count=ThreadLimit.DEFAULT, name=None):
        self.__repr_name = f"name={name}" if name else f"id={id(self)}"
        self.logger = logging.getLogger("klio.concurrency")

        if max_thread_count is ThreadLimit.DEFAULT:
            max_thread_count = multiprocessing.cpu_count

        if max_thread_count is ThreadLimit.NONE:
            self._dummy = True
            self._semaphore = _DummySemaphore()
            self.logger.debug(f"{self} Using unlimited semaphore")

        else:
            self._dummy = False
            if callable(max_thread_count):
                max_thread_count = max_thread_count()
            self._semaphore = threading.BoundedSemaphore(max_thread_count)
            self.logger.debug(
                f"{self} Initial semaphore value: {self._semaphore._value}"
            )

    def _log(self, message):
        threads_available = False
        suffix_msg = ""
        if not self._dummy:
            threads_available = self._semaphore._value
            suffix_msg = f" (available threads: {threads_available})"
        self.logger.debug(f"{self} {message}{suffix_msg}")

    def acquire(self):
        """Acquire a semaphore (a thread).

        Acquiring a semaphore will activate an available thread in Beam's
        multi-threaded environment.

        If no semaphores are available, the method will block until one
        is released via :func:`ThreadLimiter.release`.
        """
        self._log("Blocked – waiting on semaphore for an available thread")
        self._semaphore.acquire()

    def release(self):
        """Release a semaphore (a thread).

        Raises:
            `ValueError`: if the semaphore is released too many times (more
                than the value of ``ThreadLimiter._semaphore._value``).
        """
        self._semaphore.release()
        self._log("Released semaphore")

    def __repr__(self):
        return f"{self._PREFIX}({self.__repr_name})"

    def __enter__(self):
        self.acquire()
        return self._semaphore

    def __exit__(self, type, value, traceback):
        self.release()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import collections
import logging

import apache_beam as beam

from apache_beam import pvalue

from klio_core.proto import klio_pb2

from klio import utils as kutils
from klio.message import serializer
from klio.transforms import _helpers
from klio.transforms import decorators
from klio.transforms import io as io_transforms


_StubDataConfig = collections.namedtuple("_StubDataConfig", ["ping", "force"])
"""Stub data config when helpers are used without job_config.data configured."""


class KlioGcsCheckInputExists(
    _helpers._KlioInputDataMixin, _helpers._KlioGcsCheckExistsBase
):
    """Klio transform to check input exists in GCS."""

    pass


class KlioGcsCheckOutputExists(
    _helpers._KlioOutputDataMixin, _helpers._KlioGcsCheckExistsBase
):
    """Klio transform to check output exists in GCS."""

    pass


class KlioFilterPing(
    _helpers._KlioInputDataMixin, _helpers._KlioBaseDataExistenceCheck
):
    """Klio transform to tag outputs if in ping mode or not."""

    stub_config = _StubDataConfig(ping=False, force=False)

    @property
    def _data_config(self):
        try:
            return super()._data_config
        except _helpers.KlioConfigRuntimeError:
            # This error happens when users don't have anything configured in
            # `job_config.data.inputs`.
            # We want to allow folks to use this transform without needing
            # to unnecessarily define anything for `job_config.data.inputs`
            # so we just return a stub config with the defaults set.
            return self.stub_config

    def ping(self, kmsg):
        global_ping = self._data_config.ping
        msg_ping = kmsg.metadata.ping
        return msg_ping if msg_ping else global_ping

    def process(self, kmsg):
        tagged_state = _helpers.TaggedStates.DEFAULT
        item = kmsg.data.element.decode("utf-8")

        if self.ping(kmsg):
            self._klio.logger.info("Pass through '%s': Ping mode ON." % item)
            tagged_state = _helpers.TaggedStates.PASS_THRU

        else:
            self._klio.logger.debug("Process '%s': Ping mode OFF." % item)
            tagged_state = _helpers.TaggedStates.PROCESS

        yield pvalue.TaggedOutput(tagged_state.value, kmsg.SerializeToString())


class KlioFilterForce(
    _helpers._KlioOutputDataMixin, _helpers._KlioBaseDataExistenceCheck
):
    """Klio transform to tag outputs if in force mode or not."""

    stub_config = _StubDataConfig(ping=False, force=False)

    @property
    def _data_config(self):
        try:
            return super()._data_config
        except _helpers.KlioConfigRuntimeError:
            # This error happens when users don't have anything configured in
            # `job_config.data.outputs`.
            # We want to allow folks to use this transform without needing
            # to unnecessarily define anything for `job_config.data.outputs`
            # so we just return a stub config with the defaults set.
            return self.stub_config

    def force(self, kmsg):
        global_force = self._data_config.force
        msg_force = kmsg.metadata.force
        return msg_force if msg_force else global_force

    def process(self, kmsg):
        tagged_state = _helpers.TaggedStates.DEFAULT
        item_path = self._get_absolute_path(kmsg.data.element)
        item = kmsg.data.element.decode("utf-8")

        if not self.force(kmsg):
            self._klio.logger.info(
                "Pass through '%s': Force mode OFF with output found at '%s'."
                % (item, item_path)
            )
            tagged_state = _helpers.TaggedStates.PASS_THRU

        else:
            self._klio.logger.info(
                "Process '%s': Force mode ON with output found at '%s'."
                % (item, item_path)
            )
            tagged_state = _helpers.TaggedStates.PROCESS

        yield pvalue.TaggedOutput(tagged_state.value, kmsg.SerializeToString())


class KlioWriteToEventOutput(beam.PTransform):
    """Klio composite transform to write to the configured event output."""

    # NOTE: hopefully we don't get an dict lookup errors since KlioConfig
    # should raise if given an unsupported event IO transform
    CONFNAME_TO_OUT_TRANSFORM = {
        "file": io_transforms.KlioWriteToText,
        "pubsub": beam.io.WriteToPubSub,
    }

    @property
    @decorators._set_klio_context
    def _event_config(self):
        # TODO: figure out how to support multiple outputs

        if len(self._klio.config.job_config.events.outputs) > 1:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise RuntimeError(
                "The `klio.transforms.helpers.KlioWriteToEventOutput` "
                "transform does not support multiple outputs configured in "
                "`klio-job.yaml::job_config.events.outputs`."
            )

        if len(self._klio.config.job_config.events.outputs) == 0:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise RuntimeError(
                "The `klio.transforms.helpers.KlioWriteToEventOutput` "
                "requires an event output to be configured in "
                "`klio-job.yaml::job_config.events.outputs`."
            )
        return self._klio.config.job_config.events.outputs[0]

    def expand(self, pcoll):
        transform_kls = self.CONFNAME_TO_OUT_TRANSFORM[self._event_config.name]

        kwargs = self._event_config.as_dict()
        return (
            pcoll
            | "Writing to '%s'" % self._event_config.name
            >> transform_kls(**kwargs)
        )


# NOTE: not doing much right now, but allows us to extend if need be
class KlioDrop(beam.DoFn, metaclass=_helpers._KlioBaseDoFnMetaclass):
    """Klio DoFn to log & drop a KlioMessage."""

    WITH_OUTPUTS = False

    @decorators._handle_klio(max_thread_count=kutils.ThreadLimit.NONE)
    def process(self, kmsg):
        self._klio.logger.info(
            "Dropping KlioMessage - can not process '%s' any further."
            % kmsg.element
        )
        return


# TODO: this should only be temporary and removed once v2 migration is done
class _KlioTagMessageVersion(
    beam.DoFn, metaclass=_helpers._KlioBaseDoFnMetaclass
):
    WITH_OUTPUTS = True

    @decorators._set_klio_context
    def process(self, klio_message):
        # In batch, the read transform produces a KlioMessage. However, in
        # streaming, it's still bytes. And for some reason this isn't
        # pickleable when it's in its own transform.
        # TODO: maybe create a read/write klio pub/sub transform to do
        # this for us.
        if not isinstance(klio_message, klio_pb2.KlioMessage):
            klio_message = serializer.to_klio_message(
                klio_message, self._klio.config, self._klio.logger
            )

        if klio_message.version == klio_pb2.Version.V2:
            yield pvalue.TaggedOutput("v2", klio_message.SerializeToString())
        else:
            yield pvalue.TaggedOutput("v1", klio_message.SerializeToString())


# TODO: this should only be temporary and removed once v2 migration is done
class _KlioV1CheckRecipients(
    beam.DoFn, metaclass=_helpers._KlioBaseDoFnMetaclass
):
    """Check if current job should handle a received v1 message."""

    WITH_OUTPUTS = True

    @decorators._set_klio_context
    def _should_process(self, klio_message):
        downstream = klio_message.metadata.downstream
        if not downstream:
            # if there's nothing in downstream, then it means the message is
            # in top-down mode and should be handled
            return True

        current_job = klio_pb2.KlioJob()
        current_job.ParseFromString(self._klio.job)

        if _helpers._job_in_jobs(current_job, downstream):
            return True

        self._klio.logger.info(
            "Dropping KlioMessage - job not an intended recipient for message "
            "with entity_id {}.".format(klio_message.data.entity_id)
        )
        return False

    @decorators._set_klio_context
    def process(self, raw_message):
        klio_message = serializer.to_klio_message(
            raw_message, self._klio.config, self._klio.logger
        )
        if self._should_process(klio_message):
            yield pvalue.TaggedOutput(
                _helpers.TaggedStates.PROCESS.value, raw_message
            )
        else:
            yield pvalue.TaggedOutput(
                _helpers.TaggedStates.DROP.value, raw_message
            )


class KlioCheckRecipients(
    beam.DoFn, metaclass=_helpers._KlioBaseDoFnMetaclass
):
    """Check if current job should handle a received v2 message."""

    WITH_OUTPUTS = True

    @decorators._set_klio_context
    def _should_process(self, klio_message):
        intended_recipients = klio_message.metadata.intended_recipients
        # returns "anyone", "limited", or None if not set
        recipients = intended_recipients.WhichOneof("recipients")

        if recipients is None:
            # is it safe to assume if this is not set in a v2 message, it should
            # be top-down? I think this will be the case for batch
            self._klio.logger.warning(
                "Dropping KlioMessage - No 'intended_recipients' set in "
                "metadata of KlioMessage with element '{}'.".format(
                    klio_message.data.element
                )
            )
            return False

        if recipients == "anyone":
            return True

        current_job = klio_pb2.KlioJob()
        current_job.ParseFromString(self._klio.job)

        # otherwise, recipients == "limited"
        # don't process if this job is not in the intended recipients
        if not _helpers._job_in_jobs(
            current_job, intended_recipients.limited.recipients
        ):
            return False

        # if it is in the intended recipients _and_ is the job in
        # trigger_children_of, then this message was originally in top-down
        # mode, but was missing dependencies, and therefore should update the
        # message intended receipients to be "anyone" signifying top-down
        if _helpers._job_in_jobs(
            current_job, [intended_recipients.limited.trigger_children_of]
        ):
            # FYI: since 'anyone' is essentially empty msg, it can't simply
            # be assigned. To set `anyone` as the intended_recipients, use
            # kmsg.metadata.intended_recipients.anyone.SetInParent()`
            # https://stackoverflow.com/a/29651069
            intended_recipients.anyone.SetInParent()

        return True

    @decorators._set_klio_context
    def process(self, raw_message):
        klio_message = serializer.to_klio_message(
            raw_message, self._klio.config, self._klio.logger
        )
        if self._should_process(klio_message):
            # the message could have updated, so let's re-serialize to a new
            # raw message
            raw_message = klio_message.SerializeToString()
            yield pvalue.TaggedOutput(
                _helpers.TaggedStates.PROCESS.value, raw_message
            )
        else:
            yield pvalue.TaggedOutput(
                _helpers.TaggedStates.DROP.value, raw_message
            )


class KlioUpdateAuditLog(beam.DoFn, metaclass=_helpers._KlioBaseDoFnMetaclass):
    """Update a KlioMessage's audit log to include current job."""

    WITH_OUTPUTS = False

    @decorators._set_klio_context
    def _generate_current_job_object(self):
        job = klio_pb2.KlioJob()
        job.job_name = self._klio.config.job_name
        job.gcp_project = self._klio.config.pipeline_options.project
        return job

    def _create_audit_item(self):
        audit_log_item = klio_pb2.KlioJobAuditLogItem()
        audit_log_item.timestamp.GetCurrentTime()
        current_job = self._generate_current_job_object()
        audit_log_item.klio_job.CopyFrom(current_job)
        return audit_log_item

    @decorators._set_klio_context
    def process(self, raw_message):
        klio_message = serializer.to_klio_message(
            raw_message, self._klio.config, self._klio.logger
        )
        audit_log_item = self._create_audit_item()
        klio_message.metadata.job_audit_log.extend([audit_log_item])

        audit_log = klio_message.metadata.job_audit_log
        traversed_dag = " -> ".join(
            "{}::{}".format(
                str(al.klio_job.gcp_project), str(al.klio_job.job_name)
            )
            for al in audit_log
        )
        traversed_dag = "{} (current job)".format(traversed_dag)

        base_log_msg = "KlioMessage full audit log"
        log_msg = "{} - Entity ID: {} - Path: {}".format(
            base_log_msg, klio_message.data.entity_id, traversed_dag
        )
        self._klio.logger.debug(log_msg)
        yield klio_message.SerializeToString()


class KlioDebugMessage(beam.PTransform):
    """Log KlioMessage.

    Args:
        prefix (str): logging prefix. Default: ``"DEBUG"``.
        log_level (str or int): The desired log level for the KlioMessage
            logs. See `available log levels <https://docs.python.org/3/library/
            logging.html#levels>`_ for what's supported. Default: ``"INFO"``.
    """

    def __init__(self, prefix="DEBUG: ", log_level="INFO"):
        super().__init__()
        self.prefix = prefix
        self.log_level = self._get_log_level(log_level)

    def _get_log_level(self, log_level):
        # TODO: should prob do some pre-emptive checking
        if isinstance(log_level, str):
            return getattr(logging, log_level.upper())
        if isinstance(log_level, int):
            return log_level
        raise SystemExit("Unrecognized `log_level` for `KlioDebugMessage`.")

    @decorators._set_klio_context
    def print_debug(self, raw_message):
        klio_message = serializer.to_klio_message(
            raw_message, self._klio.config, self._klio.logger
        )
        self._klio.logger.log(
            self.log_level, "{}{}".format(self.prefix, klio_message)
        )
        return raw_message

    def expand(self, pipeline):
        return pipeline | beam.Map(self.print_debug)


class KlioSetTrace(beam.PTransform):
    """Insert a Python debugger trace point."""

    def set_trace(self, raw_message):
        import pdb  # don't import this at top level, just when it's needed

        pdb.set_trace()
        return raw_message

    def expand(self, pipeline):
        return pipeline | beam.Map(self.set_trace)


class KlioTriggerUpstream(beam.PTransform):
    """Trigger upstream job from current job with a given ``KlioMessage``.

    This transform will update the intended recipients in ``KlioMessage.
    metadata`` in order to trigger a partial :ref:`bottom-up execution
    <bottom-up>` of the overall graph of jobs. It will also generate a log
    message (optional), then publish the ``KlioMessage`` to the upstream's
    Pub/Sub topic.

    .. caution::

        Klio does not automatically trigger upstream jobs if input data does
        not exist. It must be used manually within a job's pipeline definition
        (in ``run.py::run``).

    .. note::

        In order to get access to input data not found, the automatic data
        existence check that Klio does must be turned off by setting
        ``klio-job.yaml::job_config.data.inputs[].skip_klio_existence_check``
        to ``True``. Then the existence check must be invoked manually. See
        example ``run.py`` and ``klio-job.yaml`` files below.

    Example usage:

    .. code-block:: python

        import apache_beam as beam
        from klio.transforms import helpers
        import transforms

        def run(input_pcol, config):
            # use the default helper transform to do the default input check
            # in order to access the output tagged with `not_found`
            input_data = input_pcol | helpers.KlioGcsCheckInputExists()

            # Pipe the input data that was not found (using Tagged Outputs)
            # into `KlioTriggerUpstream` in order to update the KlioMessage
            # metadata, log it, then publish to upstream's
            _ = input_data.not_found | helpers.KlioTriggerUpstream(
                upstream_job_name="my-upstream-job",
                upstream_topic="projects/my-gcp-project/topics/upstream-topic",
                log_level="DEBUG",
            )

            # pipe the found input pcollection into transform(s) as needed
            output_pcol = input_data.found | beam.ParDo(MyTransform())
            return output_pcol

    .. code-block:: yaml
        :emphasize-lines: 7,23

        # Example klio-job.yaml
        version: 2
        job_name: my-job
        pipeline_options:
          project: my-gcp-project
          # `KlioTriggerUpstream` only supports streaming jobs
          streaming: True
          # <-- snip -->
        job_config:
          events:
            inputs:
              - type: pubsub
                topic: projects/my-gcp-project/topics/upstream-topic-output
                subscription: projects/my-gcp-project/subscriptions/my-job-in
            # <-- snip -->
          data:
            inputs:
              - type: gcs
                location: gs://my-gcp-project/upstream-output-data
                file_suffix: .ogg
                # Be sure to skip Klio's default input existence check in
                # order to access the input data that was not found.
                skip_klio_existence_check: True

    Args:
        upstream_job_name (str): Name of upstream job.
        upstream_topic (str): Pub/Sub topic for the upstream job, in the
            form of ``project/<GCP_PROJECT>/topics/<TOPIC_NAME>``.
        log_level (str, int, or None): The desired log level for log message,
            or ``None`` if no logging is desired. See `available log levels
            <https://docs.python.org/3/library/logging.html#levels>`_ for
            what's supported. Default: ``"INFO"``.

    Raises:
        SystemExit: If the current job is not in streaming mode (set
            in `klio-job.yaml::pipeline_options.streaming`), if the
            provided log level is not recognized, or if the provided
            upstream topic is not in the correct form.
    """

    @decorators._set_klio_context
    def __init__(self, upstream_job_name, upstream_topic, log_level="INFO"):
        if self._klio.config.pipeline_options.streaming is False:
            # Fail early
            self._klio.logger.error(
                "The `KlioTriggerUpstreams` transform is only available for "
                "jobs in streaming mode."
            )
            raise SystemExit(1)

        self.upstream_job_name = upstream_job_name
        self.upstream_topic = upstream_topic
        self.upstream_gcp_project = self._get_project_from_topic()
        self.log_level = self._get_log_level(log_level)

    def _get_project_from_topic(self):
        stems = self.upstream_topic.split("/")
        if len(stems) != 4:
            # Fail early
            self._klio.logger.error(
                "The provided upstream topic for `KlioTriggerUpstream` is "
                "expected in the form of 'project/<GCP_PROJECT>/topics/"
                "<TOPIC_NAME>'. Received '{}'.".format(self.upstream_topic)
            )
            raise SystemExit(1)
        return stems[1]

    def _get_log_level(self, log_level):
        if log_level is None:
            return log_level

        if isinstance(log_level, (str, int)):
            # save to a different variable so we don't alter it for the error
            # log message below
            _log_level = log_level
            if isinstance(_log_level, str):
                _log_level = _log_level.upper()

            level = logging.getLevelName(_log_level)
            if isinstance(level, int):
                return level

            # getLevelName will create a level if it doesn't recognize it (wtf)
            # so let's check if it actually exists
            ret = getattr(logging, level, None)
            if ret is not None:
                return ret
        self._klio.logger.error(
            "Unrecognized log level '{}' for `KlioTriggerUpstream`.".format(
                log_level
            )
        )
        raise SystemExit(1)

    def _generate_upstream_job_object(self):
        upstream_job = klio_pb2.KlioJob()
        upstream_job.job_name = self.upstream_job_name
        upstream_job.gcp_project = self.upstream_gcp_project
        return upstream_job

    def default_label(self):
        # Will be overwritten when invoked with a custom label, i.e.
        # `"Some Label" >> KlioTriggerUpstream(...)`
        return "{}(upstream={})".format(
            self.__class__.__name__, self.upstream_job_name
        )

    @decorators._set_klio_context
    def _generate_current_job_object(self):
        job = klio_pb2.KlioJob()
        job.job_name = self._klio.config.job_name
        job.gcp_project = self._klio.config.pipeline_options.project
        return job

    @decorators._set_klio_context
    def update_kmsg_metadata(self, raw_kmsg):
        """Update KlioMessage to enable partial bottom-up execution.

        Args:
            raw_kmsg (bytes): Unserialized KlioMessage
        Returns:
            bytes: KlioMessage deserialized to ``bytes`` with updated intended
                recipients metadata.
        """
        # Use `serializer.to_klio_message` instead of @handle_klio in order to
        # get the full KlioMessage object (not just the data).
        kmsg = serializer.to_klio_message(
            raw_kmsg, kconfig=self._klio.config, logger=self._klio.logger
        )

        # Make sure upstream job doesn't skip the message
        upstream_job = self._generate_upstream_job_object()
        lmtd = kmsg.metadata.intended_recipients.limited
        lmtd.recipients.extend([upstream_job])

        # Assign the current job to `trigger_children_of` so that top-down
        # execution resumes after this job is done.
        current_job = self._generate_current_job_object()
        lmtd.recipients.extend([current_job])
        lmtd.trigger_children_of.CopyFrom(current_job)
        return serializer.from_klio_message(kmsg)

    @decorators._handle_klio(max_thread_count=kutils.ThreadLimit.NONE)
    def log(self, data):
        """Log KlioMessage being published to upstream job.

        Args:
            data (klio_core.proto.KlioMessage.data): data of ``KlioMessage``.
        Returns:
            klio_core.proto.KlioMessage.data: unchanged data of
                ``KlioMessage``.
        """
        element = data.element.decode("utf-8")
        msg = "Triggering upstream {} for {}".format(
            self.upstream_job_name, element
        )
        self._klio.logger.log(self.log_level, msg)
        return data

    def expand(self, pcoll):
        name = self.upstream_job_name
        lbl1 = "Update KlioMessage for Upstream {}".format(name)
        lbl2 = "Log Triggering Upstream {}".format(name)
        lbl3 = "Publish KlioMessage to Upstream {}".format(name)

        updated_kmsg = pcoll | lbl1 >> beam.Map(self.update_kmsg_metadata)

        if self.log_level is not None:
            _ = updated_kmsg | lbl2 >> beam.Map(self.log)

        return updated_kmsg | lbl3 >> beam.io.WriteToPubSub(
            topic=self.upstream_topic
        )



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio.metrics import client
from klio.metrics import dispatcher


@pytest.fixture
def metric_params():
    return {
        "name": "my-metric",
        "value": 0,
        "kwargs": {"tags": {"key-tag": "value-tag"}},
    }


@pytest.fixture
def relay_client(mocker):
    return mocker.Mock()


@pytest.fixture
def metrics_registry(relay_client):
    return client.MetricsRegistry(
        relay_clients=[relay_client], transform_name="HelloKlio"
    )


@pytest.fixture
def metric_data(metric_params):
    metric_params["type"] = "counter"
    return metric_params


@pytest.fixture
def counter_metric(metric_params, relay_client):
    return dispatcher.CounterDispatcher(
        relay_clients=[relay_client], **metric_params
    )


@pytest.mark.parametrize(
    "method,cls",
    (
        ("counter", dispatcher.CounterDispatcher),
        ("gauge", dispatcher.GaugeDispatcher),
        ("timer", dispatcher.TimerDispatcher),
    ),
)
def test_get_metric_inst(method, cls, metrics_registry, metric_params):
    assert {} == metrics_registry._registry  # sanity check

    method_to_call = getattr(metrics_registry, method)
    ret_metric = method_to_call(**metric_params)

    assert isinstance(ret_metric, cls)

    exp_key = "{method}_{name}_HelloKlio".format(
        method=method, **metric_params
    )
    assert exp_key in metrics_registry._registry
    assert ret_metric == metrics_registry._registry[exp_key]

    # assert same metric is returned rather than creating a new instance
    ret_metric_again = method_to_call(**metric_params)
    assert ret_metric is ret_metric_again


@pytest.mark.parametrize(
    "metric_type,cls",
    (
        ("counter", dispatcher.CounterDispatcher),
        ("gauge", dispatcher.GaugeDispatcher),
        ("timer", dispatcher.TimerDispatcher),
        ("unknown", dispatcher.GaugeDispatcher),
    ),
)
def test_marshal_unmarshal(
    metrics_registry, metric_type, cls, metric_params, relay_client
):
    metric_inst = cls(
        relay_clients=[relay_client], transform="HelloKlio", **metric_params
    )
    metric_data = metric_params.copy()
    metric_data["type"] = metric_type
    ret_metric_data = metrics_registry.marshal(metric_inst)

    exp_metric_data = metric_data.copy()
    if metric_type == "unknown":
        exp_metric_data["type"] = "gauge"
    if metric_type == "timer":
        exp_metric_data["timer_unit"] = "ns"
    assert exp_metric_data == ret_metric_data

    ret_metric = metrics_registry.unmarshal(metric_data)

    # can't simply compare exp_metric to ret_metric since they are
    # two different instances
    # FIXME: implement __eq__ and __ne__ for dispatch objects (@lynn)
    assert metric_inst.METRIC_TYPE == ret_metric.METRIC_TYPE
    assert metric_inst.name == ret_metric.name
    assert metric_inst.value == ret_metric.value
    assert metric_inst.transform == ret_metric.transform
    assert metric_inst.kwargs == ret_metric.kwargs



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Klio ships with a default :class:`klio.metrics.base.AbstractRelayClient`
implementation, which outputs metrics via the standard library ``logging``
module through the :class:`MetricsLoggerClient` below.

This implementation is used by default if no other metrics consumers are
configured. It must be explicitly turned off.

The default configuration in ``klio-info.yaml`` can be overwritten:

.. code-block:: yaml

    job_config:
        metrics:
            logger:
                # Logged metrics are emitted at the `debug` level by default.
                level: info
                # Default timer unit is ns/nanoseconds; available
                # options include `s` or `seconds`, `ms` or `milliseconds`,
                # `us` or `microseconds`, and `ns` or `nanoseconds`.
                timer_unit: s

To turn off logging-based metrics:

.. code-block:: yaml

    job_config
        metrics:
            logger: false
"""

import logging
import threading

from klio.metrics import base


TIMER_UNIT_MAP = {
    "nanoseconds": "ns",
    "microseconds": "us",
    "milliseconds": "ms",
    "seconds": "s",
    "ns": "ns",
    "us": "us",
    "ms": "ms",
    "s": "s",
}
"""Map of supported measurement units to shorthand for :class:`LoggerTimer`.
"""


class MetricsLoggerClient(base.AbstractRelayClient):
    """Logging client for transform metrics.

    Intended to be instantiated by :class:`klio.metrics.client.MetricsRegistry`
    and not by itself.

    Args:
        klio_config (klio_core.config.KlioConfig):  the job's configuration.
        disabled (bool): whether or not to disable the Python ``logger``
            Default: ``False``.
    """

    RELAY_CLIENT_NAME = "logger"
    DEFAULT_LEVEL = logging.DEBUG
    DEFAULT_TIME_UNIT = "ns"

    _thread_local = threading.local()

    def __init__(self, klio_config, disabled=False):
        super(MetricsLoggerClient, self).__init__(klio_config)
        self.logger_config = self.klio_config.job_config.metrics.get(
            "logger", {}
        )
        self.disabled = disabled
        self.log_level = self._set_log_level()
        self.timer_unit = self._set_timer_unit()

    def _set_log_level(self):
        log_level = MetricsLoggerClient.DEFAULT_LEVEL
        if isinstance(self.logger_config, dict):
            log_level_str = self.logger_config.get("level")
            if log_level_str:
                log_level = getattr(logging, log_level_str.upper(), log_level)
        return log_level

    def _set_timer_unit(self):
        timer_unit = MetricsLoggerClient.DEFAULT_TIME_UNIT
        if isinstance(self.logger_config, dict):
            _timer_unit = self.logger_config.get("timer_unit")
            if _timer_unit:
                timer_unit = TIMER_UNIT_MAP.get(_timer_unit, timer_unit)
        return timer_unit

    @property
    def logger(self):
        """Python logger associated with the job which this client will use
        to emit metrics.
        """
        klio_metrics_logger = getattr(
            self._thread_local, "klio_metrics_logger", None
        )
        if not klio_metrics_logger:
            logger = logging.getLogger("klio.metrics")
            logger.disabled = self.disabled
            self._thread_local.klio_metrics_logger = logger
        return self._thread_local.klio_metrics_logger

    def unmarshal(self, metric):
        """Return a dict-representation of a given metric.

        Args:
            metric (LoggerMetric): logger-specific metrics object
        Returns:
            dict(str, str): metric data
        """
        return {
            "name": metric.name,
            "value": metric.value,
            "transform": metric.transform,
            "tags": metric.tags,
        }

    def emit(self, metric):
        """Log a given metric.

        Args:
            metric (LoggerMetric): logger-specific metrics object
        """
        metric_data = self.unmarshal(metric)
        self.logger.log(
            self.log_level, metric.DEFAULT_LOG_FORMAT.format(**metric_data)
        )

    def counter(self, name, value=0, transform=None, tags=None, **kwargs):
        """Create a :class:`LoggerCounter` object.

        Args:
            name (str): name of counter
            value (int): starting value of counter; defaults to 0
            transform (str): transform the counter is associated with
            tags (dict): any tags of additional contextual information
                to associate with the counter

        Returns:
            LoggerCounter: a log-based counter
        """
        return LoggerCounter(
            name=name, value=value, transform=transform, tags=tags
        )

    def gauge(self, name, value=0, transform=None, tags=None, **kwargs):
        """Create a :class:`LoggerGauge` object.

        Args:
            name (str): name of gauge
            value (int): starting value of gauge; defaults to 0
            transform (str): transform the gauge is associated with
            tags (dict): any tags of additional contextual information
                to associate with the gauge

        Returns:
            LoggerGauge: a log-based gauge
        """
        return LoggerGauge(
            name=name, value=value, transform=transform, tags=tags
        )

    def timer(
        self,
        name,
        value=0,
        transform=None,
        tags=None,
        timer_unit=None,
        **kwargs
    ):
        """Create a :class:`LoggerTimer` object.

        Args:
            name (str): name of timer
            value (int): starting value of timer; defaults to 0
            transform (str): transform the timer is associated with
            tags (dict): any tags of additional contextual information
                to associate with the timer
            timer_unit (str): timer unit; defaults to configured value
                in `klio-job.yaml`, or "ns". See module-level docs of
                `klio.metrics.logger` for supported values.

        Returns:
            LoggerTimer: a log-based timer
        """
        if timer_unit:
            # Note: this should probably have better validation if it does
            # not recognize the unit given. Instead of erroring out, we'll
            # just use the default (@lynn)
            timer_unit = TIMER_UNIT_MAP.get(timer_unit, self.timer_unit)
        else:
            timer_unit = self.timer_unit
        return LoggerTimer(
            name=name,
            value=value,
            transform=transform,
            tags=tags,
            timer_unit=timer_unit,
        )


class LoggerMetric(base.BaseMetric):
    """Base metric type for loggers.

    Args:
        name (str): name of counter
        value (int): initial value. Default: ``0``.
        transform (str): Name of transform associated with metric, if any.
        tags (dict): Tags to associate with metric.
    """

    LOGGER_METRIC_TAGS = None
    DEFAULT_LOG_FORMAT = (
        "[{name}] value: {value} transform: '{transform}' tags: {tags}"
    )

    def __init__(self, name, value=0, transform=None, tags=None):
        super(LoggerMetric, self).__init__(
            name, value=value, transform=transform
        )
        self.tags = tags if tags else {}
        self.tags.update(self.LOGGER_METRIC_TAGS)


class LoggerCounter(LoggerMetric):
    """Log-based counter metric.

    Args:
        name (str): name of counter
        value (int): initial value. Default: ``0``.
        transform (str): Name of transform associated with counter, if any.
        tags (dict): Tags to associate with counter. Note:
            ``{"metric_type": "counter"}`` will always be an included tag.
    """

    LOGGER_METRIC_TAGS = {"metric_type": "counter"}


class LoggerGauge(LoggerMetric):
    """Log-based gauge metric.

    Args:
        name (str): name of gauge
        value (int): initial value. Default: ``0``.
        transform (str): Name of transform associated with gauge, if any.
        tags (dict): Tags to associate with gauge. Note:
            ``{"metric_type": "gauge"}`` will always be an included tag.
    """

    LOGGER_METRIC_TAGS = {"metric_type": "gauge"}


class LoggerTimer(LoggerMetric):
    """Log-based timer metric.

    Args:
        name (str): name of timer
        value (int): initial value. Default: ``0``.
        transform (str): Name of transform associated with timer, if any.
        tags (dict): Tags to associate with timer. Note:
            ``{"metric_type": "timer"}`` will always be an included tag.
        timer_unit (str): Unit of measurement. Options: :attr:`TIMER_UNIT_MAP`.
            Default: ``ns`` (nanoseconds).
    """

    LOGGER_METRIC_TAGS = {"metric_type": "timer"}

    def __init__(
        self, name, value=0, transform=None, tags=None, timer_unit="ns"
    ):
        self.LOGGER_METRIC_TAGS.update({"unit": timer_unit})
        super(LoggerTimer, self).__init__(
            name, value=value, transform=transform, tags=tags
        )
        self.timer_unit = timer_unit



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import __main__
import glob
import logging
import os
import threading

import yaml

from klio_core import config
from klio_core.config import core as config_core
from klio_core.proto import klio_pb2

from klio.metrics import client as metrics_client
from klio.metrics import logger as metrics_logger
from klio.metrics import stackdriver


class RunConfig(object):

    _thread_local = threading.local()

    @classmethod
    def _load_config_from_file(cls):
        klio_job_file = None

        if os.path.exists(config_core.WORKER_RUN_EFFECTIVE_CONFIG_PATH):
            klio_job_file = config_core.WORKER_RUN_EFFECTIVE_CONFIG_PATH
        else:
            run_config_path = os.path.join(
                "/usr/**", config_core.RUN_EFFECTIVE_CONFIG_FILE
            )
            files = glob.iglob(run_config_path, recursive=True)
            for f in files:
                klio_job_file = f
                # only grab the first one
                break

        if not klio_job_file:
            klio_job_file = "/usr/src/config/.effective-klio-job.yaml"

        logger = logging.getLogger("klio")
        logger.debug(f"Loading config file from {klio_job_file}.")

        with open(klio_job_file, "r") as f:
            all_config_data = yaml.safe_load(f)
        return config.KlioConfig(all_config_data)

    # NOTE: for now this approach is not being used (and may be removed in the
    # future)
    @classmethod
    def _get_via_main_session(cls):
        if hasattr(__main__, "run_config"):
            return __main__.run_config
        else:
            raise Exception(
                "Attempt to access RunConfig before it was set. This likely"
                " means something was imported before RunConfig was set."
            )

    @classmethod
    def _get_via_thread_local(cls):
        klio_config = getattr(cls._thread_local, "klio_config", None)
        if not klio_config:
            cls._thread_local.klio_config = cls._load_config_from_file()
        return cls._thread_local.klio_config

    @classmethod
    def get(cls):
        return cls._get_via_thread_local()

    @classmethod
    def set(cls, config):
        __main__.run_config = config


class KlioContext(object):
    """Context related to the currently running job.

    Available to transforms via one of the :ref:`KlioContext decorators
    <klio-context-decorators>`.
    """

    _thread_local = threading.local()

    def __init__(self):
        self.__transform_name = None

    def _create_klio_job_obj(self):
        klio_job = klio_pb2.KlioJob()
        klio_job.job_name = self.config.job_name
        klio_job.gcp_project = self.config.pipeline_options.project
        klio_job_str = klio_job.SerializeToString()
        return klio_job_str

    def _get_metrics_registry(self):
        clients = []
        use_logger, use_stackdriver = None, None
        metrics_config = self.config.job_config.metrics

        # use_logger and use_stackdriver could be False (turn off),
        # None (use default config), or a dict of configured values
        use_logger = metrics_config.get("logger")
        use_stackdriver = metrics_config.get("stackdriver_logger")

        # TODO: set runner in OS environment (via klio-exec), since
        #       the runner defined in config could be overwritten via
        #       `--direct-runner`.
        #       i.e.: runner = os.getenv("BEAM_RUNNER", "").lower()
        runner = self.config.pipeline_options.runner
        if "dataflow" in runner.lower():
            # Must explicitly compare to `False` since `None` could be
            # the user accepting default config.
            # If explicitly false, then just disable logger underneath SD
            if use_stackdriver is not False:
                sd_client = stackdriver.StackdriverLogMetricsClient(
                    self.config
                )
                clients.append(sd_client)
            else:
                # if use_stackdriver is explicitly false, then make sure
                # logger client is disabled since the stackdriver client
                # inherits the logger client
                use_logger = False

        if not len(clients):  # setup default client
            disabled = False
            # User might disable the logger, but we still need a relay
            # client if all other relay clients are disabled. This allows
            # folks to silence metrics but not need to remove code that
            # interacts with `_klio.metrics`.
            # Must explicitly compare to `False` since `None` could be
            # the user accepting default config
            if use_logger is False:
                disabled = True
            logger_client = metrics_logger.MetricsLoggerClient(
                self.config, disabled=disabled
            )
            clients.append(logger_client)

        return metrics_client.MetricsRegistry(
            clients, transform_name=self._transform_name
        )

    @property
    def config(self):
        """A ``KlioConfig`` instance representing the job's configuration."""
        return RunConfig.get()

    @property
    def job(self):
        """An instance of :ref:`kliojob` of the current job."""
        klio_job = getattr(self._thread_local, "klio_job", None)
        if not klio_job:
            self._thread_local.klio_job = self._create_klio_job_obj()
        return self._thread_local.klio_job

    @property
    def logger(self):
        """A namespaced logger.

        Equivalent to ``logging.getLogger("klio")``.
        """
        klio_logger = getattr(self._thread_local, "klio_logger", None)
        if not klio_logger:
            self._thread_local.klio_logger = logging.getLogger("klio")
        return self._thread_local.klio_logger

    @property
    def metrics(self):
        """A metrics registry instance.

        See :ref:`metrics <metrics>` for more information."""
        metrics_registry = getattr(self._thread_local, "klio_metrics", None)
        if not metrics_registry:
            self._thread_local.klio_metrics = self._get_metrics_registry()
        return self._thread_local.klio_metrics

    # <-- private/internal attributes -->
    @property
    def _transform_name(self):
        return self.__transform_name

    @_transform_name.setter
    def _transform_name(self, name):
        self.__transform_name = name



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio.transforms import _helpers
from klio.transforms import _utils


class BaseKlassMeta(object):
    __metaclass__ = _helpers._KlioBaseDataExistenceCheck


class KlassMeta(BaseKlassMeta):
    def process(self):
        pass

    def not_process(self):
        pass


@pytest.mark.parametrize(
    "clsdict,bases,expected",
    (
        (
            {"process": KlassMeta.process},
            (_helpers._KlioBaseDataExistenceCheck,),
            True,
        ),
        (
            {"process": KlassMeta.not_process},
            (_helpers._KlioBaseDataExistenceCheck,),
            False,
        ),
        (
            {"process": "not a callable"},
            (_helpers._KlioBaseDataExistenceCheck,),
            False,
        ),
        ({"process": KlassMeta.process}, (BaseKlassMeta,), False),
        ({}, (BaseKlassMeta,), False),
    ),
)
def test_is_original_process_func(clsdict, bases, expected):
    actual = _utils.is_original_process_func(
        clsdict, bases, base_class="_KlioBaseDataExistenceCheck"
    )

    assert expected == actual



#! /usr/bin/env python
#
# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import codecs
import os
import re

from setuptools import find_packages
from setuptools import setup


HERE = os.path.abspath(os.path.dirname(__file__))
ROOT_DIR = os.path.abspath(os.path.join(HERE, os.path.pardir))


#####
# Helper functions
#####
def read(*filenames, **kwargs):
    """
    Build an absolute path from ``*filenames``, and  return contents of
    resulting file.  Defaults to UTF-8 encoding.
    """
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for fl in filenames:
        with codecs.open(os.path.join(HERE, fl), "rb", encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    re_str = r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta)
    meta_match = re.search(re_str, META_FILE, re.M)
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


def get_recent_changelog(filename):
    """Grab the entry for the latest release."""
    contents = read(filename).split("\n")
    # Skip the title (first two lines)
    contents = contents[2:]

    # Find the latest release
    start_line, end_line = None, None
    for i, line in enumerate(contents):
        # if we've found the start & end, we're done
        if all([start_line, end_line]):
            break

        # 4 dashes are a horizontal line in rST, so look for headers with 5+
        if line.startswith("-----"):
            if start_line is None:
                start_line = i - 1
                continue
            if end_line is None:
                end_line = i - 1
                continue

    recent_log_lines = contents[start_line:end_line]
    recent_log = "\n".join(recent_log_lines)
    return recent_log


def get_main_readme_intro():
    """Grab the intro from the README from the root of the project."""
    readme_path = os.path.join(ROOT_DIR, "README.rst")
    contents = read(readme_path).split(".. start-long-desc")[1:]
    return "\n".join(contents)


def get_long_description(package_dir):
    """Generate the long description from the README and changelog."""
    cl_base_path = f"reference/{package_dir}/changelog"
    cl_file_path = os.path.join(ROOT_DIR, f"docs/src/{cl_base_path}.rst")
    cl_url = f"{PROJECT_URLS['Documentation']}/en/latest/{cl_base_path}.html"

    # When tox builds the library, it loses where the ROOT_DIR is; this
    # is also a problem with our integration tests. So we'll just ignore
    # if there's an error and continue on
    main_readme, recent_changelog = "", ""
    try:
        main_readme = get_main_readme_intro()
        recent_changelog = get_recent_changelog(cl_file_path)
    except Exception:
        pass

    desc = (
        read("README.rst")
        + "\n"
        + main_readme
        + "\n"
        + "Release Information\n"
        + "===================\n\n"
        + recent_changelog
        + f"\n\n`Full Changelog <{cl_url}>`_.\n\n"
    )
    return desc


#####
# Project-specific constants
#####
NAME = "klio"
PACKAGE_NAME = "klio"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", PACKAGE_NAME, "__init__.py")
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS :: MacOS X",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "License :: OSI Approved :: Apache Software License",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Multimedia",
    "Topic :: Multimedia :: Graphics",
    "Topic :: Multimedia :: Sound/Audio",
    "Topic :: Multimedia :: Video",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Image Processing",
]
KEYWORDS = ["klio", "apache", "beam", "audio"]
PROJECT_URLS = {
    "Documentation": "https://docs.klio.io",
    "Bug Tracker": "https://github.com/spotify/klio/issues",
    "Source Code": "https://github.com/spotify/klio"
}
META_FILE = read(META_PATH)
INSTALL_REQUIRES = [
    # 2.22 added DirectRunner support for `DoFn.setup`
    "apache-beam[gcp]>2.21.0",
    "google-api-python-client",
    # TODO: change this to >=0.2.3 on final release
    "klio-core>=0.2.3rc1",
    "protobuf",
    "psutil",
    "pyyaml",
    "six",
]
EXTRAS_REQUIRE = {
    "docs": ["sphinx", "interrogate"],
    "tests": [
        "coverage",
        "pytest>=4.3.0",  # 4.3.0 dropped last use of `convert`
        "pytest-cov",
        "pytest-mock",
    ],
    "audio": ["klio-audio"]
}
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["tests"] + ["bumpversion", "wheel"]
)
# support 3.6, 3.7, & 3.8, matching Beam's support
PYTHON_REQUIRES = ">=3.6, <3.9"

setup(
    name=NAME,
    version=find_meta("version"),
    description=find_meta("description"),
    long_description=get_long_description("lib"),
    long_description_content_type="text/x-rst",
    url=find_meta("uri"),
    project_urls=PROJECT_URLS,
    keywords=KEYWORDS,
    author=find_meta("author"),
    author_email=find_meta("email"),
    maintainer=find_meta("author"),
    maintainer_email=find_meta("email"),
    packages=PACKAGES,
    package_dir={"": "src"},
    include_package_data=True,
    classifiers=CLASSIFIERS,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=PYTHON_REQUIRES,
)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Metric dispatchers for each metric type.

A dispatcher provides one single metric instance that will then interface
with all configured relay clients for that particular metric time. For
example:

.. code-block:: python

    relay_clients = [
        klio.metrics.logger.MetricsLoggerClient(config),
        some_other_relay_client,
    ]
    my_counter = CounterDispatcher(relay_clients, name="my-counter")
    my_counter.inc()

Creating the ``my_counter`` instance will create two relay counter
instances, each specific to each of the relay clients configured.
Calling ``inc()`` on ``my_counter`` will then call ``emit`` on each relay
counter instance where each relay client will take care of its own
emit logic.
"""

import logging
import threading
import timeit

from concurrent import futures


# not making it an ABC because there shouldn't be a need to create custom
# dispatchers
class BaseMetricDispatcher(object):
    """Base class for metric-specific dispatching.

    Each type of metric (counter, gauge, timer) requires a dispatcher
    implementation.
    """

    METRIC_TYPE = None
    _thread_local = threading.local()

    def __init__(self, relay_clients, name, value=0, transform=None, **kwargs):
        self.name = name
        self.value = value
        self.transform = transform
        self.metric_key = self._setup_metric_key()
        self.kwargs = kwargs
        self.relay_to_metric = self._setup_metric_relay(relay_clients)
        self._thread_pool = futures.ThreadPoolExecutor(
            max_workers=len(relay_clients)
        )

    def _setup_metric_relay(self, relay_clients):
        raise NotImplementedError()

    def _setup_metric_key(self):
        metric_key = "{}_{}".format(self.METRIC_TYPE, self.name)
        if self.transform:
            metric_key = "{}_{}".format(metric_key, self.transform)
        return metric_key

    @property
    def logger(self):
        """Python logger associated with metric dispatcher."""
        klio_metrics_dispatcher_logger = getattr(
            self._thread_local, "klio_metrics_dispatcher_logger", None
        )
        if not klio_metrics_dispatcher_logger:
            logger = logging.getLogger("klio.metrics.dispatcher")
            self._thread_local.klio_metrics_dispatcher_logger = logger
        return self._thread_local.klio_metrics_dispatcher_logger

    def _submit_callback(self, fut):
        try:
            fut.result()
            # no need to do anything if successful
        except Exception as e:
            msg = "Error emitting metric '{}': {}".format(fut.metric_key, e)
            self.logger.warning(msg)

    def submit(self, emit, metric):
        """Emit metrics via a threadpool."""
        fut = self._thread_pool.submit(emit, metric)

        fut.metric_key = self.metric_key  # for easy identifying
        fut.add_done_callback(self._submit_callback)


class CounterDispatcher(BaseMetricDispatcher):
    """Counter-like object that will emit via all configured clients."""

    METRIC_TYPE = "counter"

    def _setup_metric_relay(self, relay_clients):
        return [
            (
                r,
                r.counter(
                    name=self.name,
                    value=self.value,
                    transform=self.transform,
                    **self.kwargs
                ),
            )
            for r in relay_clients
        ]

    def inc(self, value=1):
        """Increment counter.

        Calling this method will emit the metric via configured clients.

        Args:
            value (int): value with which to increment the counter;
                default is 1.
        """
        self.value += value

        for relay, counter in self.relay_to_metric:
            counter.update(self.value)
            self.submit(relay.emit, counter)


class GaugeDispatcher(BaseMetricDispatcher):
    """Gauge-like object that will emit via all configured clients."""

    METRIC_TYPE = "gauge"

    def _setup_metric_relay(self, relay_clients):
        return [
            (
                r,
                r.gauge(
                    name=self.name,
                    value=self.value,
                    transform=self.transform,
                    **self.kwargs
                ),
            )
            for r in relay_clients
        ]

    def set(self, value):
        """Set gauge to a given value.

        Calling this method will emit the metric via configured clients.

        Args:
            value (int): value with which to set the gauge.
        """
        self.value = value

        for relay, gauge in self.relay_to_metric:
            gauge.update(self.value)
            self.submit(relay.emit, gauge)


class TimerDispatcher(BaseMetricDispatcher):
    """Timer-like object that will emit via all configured clients.

    This may be used by instantiating and manually calling start & stop:

    .. code-block:: python

        timer = TimerDispatcher(relay_clients, name)
        timer.start()
        # code to time
        timer.stop()

    Or as a context manager:

    .. code-block:: python

        with TimerDispatcher(relay_clients, name):
            # code to time
    """

    METRIC_TYPE = "timer"
    TIMER_UNIT_TO_NUMBER = {
        "ns": 1e9,  # seconds -> nanoseconds, default
        "us": 1e6,  # seconds -> microseconds
        "ms": 1e3,  # seconds -> milliseconds
        "s": 1,
    }

    def __init__(
        self,
        relay_clients,
        name,
        value=0,
        transform=None,
        timer_unit="ns",
        **kwargs
    ):
        super(TimerDispatcher, self).__init__(
            relay_clients=relay_clients,
            name=name,
            value=value,
            transform=transform,
            timer_unit=timer_unit,
            **kwargs
        )
        self.timer_unit = timer_unit
        self._start_time = None

    def _setup_metric_relay(self, relay_clients):
        return [
            (
                r,
                r.timer(
                    name=self.name,
                    value=self.value,
                    transform=self.transform,
                    **self.kwargs
                ),
            )
            for r in relay_clients
        ]

    def start(self):
        """Start the timer."""
        self._start_time = timeit.default_timer()

    def stop(self):
        """Stop the timer.

        Calling this method will emit the metric via configured clients.
        """
        if self._start_time is None:
            self.logger.warning(
                "Timer {} cannot be stopped before started.".format(
                    self.metric_key
                )
            )
            return

        time_elapsed = timeit.default_timer() - self._start_time
        self.value = time_elapsed * TimerDispatcher.TIMER_UNIT_TO_NUMBER.get(
            self.timer_unit, 1e9
        )

        for relay, timer in self.relay_to_metric:
            timer.update(self.value)
            self.submit(relay.emit, timer)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import glob
import json
import os
import tempfile

import apache_beam as beam
import pytest

from apache_beam.testing import test_pipeline

from klio_core.proto import klio_pb2

from klio.transforms import io as io_transforms


HERE = os.path.abspath(os.path.join(os.path.abspath(__file__), os.path.pardir))
FIXTURE_PATH = os.path.join(HERE, os.path.pardir, "fixtures")


def assert_expected_klio_msg_from_file(element):
    message = klio_pb2.KlioMessage()
    message.ParseFromString(element)
    assert message.data.element is not None
    assert isinstance(message.data.element, bytes)


def test_read_from_file():
    file_path = os.path.join(FIXTURE_PATH, "elements_text_file.txt")

    transform = io_transforms.KlioReadFromText(file_path)
    with test_pipeline.TestPipeline() as p:
        (
            p
            | "Read" >> transform
            | beam.Map(assert_expected_klio_msg_from_file)
        )

    assert transform._REQUIRES_IO_READ_WRAP is False


def test_write_to_file():
    file_path_read = os.path.join(FIXTURE_PATH, "elements_text_file.txt")

    with tempfile.TemporaryDirectory() as tmp_path:
        with test_pipeline.TestPipeline() as p:
            (
                p
                | io_transforms.KlioReadFromText(file_path_read)
                | io_transforms.KlioWriteToText(tmp_path)
            )
        # WriteToText will shard files so we iterate through each
        # file in the directory
        write_results = []

        for file_name in glob.glob(tmp_path + "*"):
            if os.path.isfile(os.path.join(tmp_path, file_name)):
                with open(file_name, "rb") as f:
                    write_results.extend(f.readlines())
        with open(file_path_read, "rb") as fr:
            read_results = fr.readlines()
        assert write_results == read_results


def _expected_avro_kmsgs():
    expected_records = [
        {
            "username": "miguno",
            "tweet": "Rock: Nerf paper, scissors is fine.",
            "timestamp": 1366150681,
        },
        {
            "username": "BlizzardCS",
            "tweet": "Works as intended.  Terran is IMBA.",
            "timestamp": 1366154481,
        },
    ]
    expected_kmsgs = []
    for record in expected_records:
        message = klio_pb2.KlioMessage()
        message.version = klio_pb2.Version.V2
        message.metadata.intended_recipients.anyone.SetInParent()
        message.data.element = bytes(json.dumps(record).encode("utf-8"))
        expected_kmsgs.append(message)
    return expected_kmsgs


def assert_expected_klio_msg_from_avro(element):
    expected_kmsgs = _expected_avro_kmsgs()
    message = klio_pb2.KlioMessage()
    message.ParseFromString(element)
    assert message in expected_kmsgs


def test_read_from_avro():
    file_pattern = os.path.join(FIXTURE_PATH, "twitter.avro")

    with test_pipeline.TestPipeline() as p:
        (
            p
            | io_transforms.KlioReadFromAvro(file_pattern=file_pattern)
            | beam.Map(assert_expected_klio_msg_from_avro)
        )

    assert io_transforms.KlioReadFromAvro._REQUIRES_IO_READ_WRAP is True


def assert_expected_klio_msg_from_avro_write(element):
    file_path_read = os.path.join(FIXTURE_PATH, "elements_text_file.txt")
    with open(file_path_read, "rb") as fr:
        expected_elements = fr.read().splitlines()
    message = klio_pb2.KlioMessage()
    message.ParseFromString(element)
    assert message.data.element in expected_elements


def test_write_to_avro():

    file_path_read = os.path.join(FIXTURE_PATH, "elements_text_file.txt")

    with tempfile.TemporaryDirectory() as tmp_path:
        with test_pipeline.TestPipeline() as p:

            p | io_transforms.KlioReadFromText(
                file_path_read
            ) | io_transforms.KlioWriteToAvro(file_path_prefix=tmp_path)

        files = glob.glob(tmp_path + "*")
        assert len(files) > 0
        assert (
            os.path.isfile(os.path.join(tmp_path, file_name))
            for file_name in files
        )

        with test_pipeline.TestPipeline() as p2:
            p2 | io_transforms.KlioReadFromAvro(
                file_pattern=(tmp_path + "*")
            ) | beam.Map(assert_expected_klio_msg_from_avro_write)


def test_avro_io_immutability():

    initial_data_path = os.path.join(FIXTURE_PATH, "twitter.avro")

    with tempfile.TemporaryDirectory() as tmp_path:
        with test_pipeline.TestPipeline() as p:

            p | io_transforms.KlioReadFromAvro(
                initial_data_path
            ) | io_transforms.KlioWriteToAvro(
                file_path_prefix=tmp_path, num_shards=0
            )

        with test_pipeline.TestPipeline() as p2:

            p2 | io_transforms.KlioReadFromAvro(
                file_pattern=tmp_path + "*"
            ) | beam.Map(assert_expected_klio_msg_from_avro)


def test_bigquery_mapper_generate_klio_message():

    mapper = io_transforms._KlioReadFromBigQueryMapper()
    message = mapper._generate_klio_message()

    assert message.version == klio_pb2.Version.V2
    assert (
        message.metadata.intended_recipients.WhichOneof("recipients")
        == "anyone"
    )


@pytest.mark.parametrize(
    "klio_message_columns,row,expected",
    (
        (["one_column"], {"a": "A", "b": "B", "one_column": "value"}, "value"),
        (
            ["a", "b"],
            {"a": "A", "b": "B", "c": "C"},
            json.dumps({"a": "A", "b": "B"}),
        ),
        (None, {"a": "A", "b": "B"}, json.dumps({"a": "A", "b": "B"})),
    ),
)
def test_bigquery_mapper_map_row_element(klio_message_columns, row, expected):
    mapper = io_transforms._KlioReadFromBigQueryMapper(
        klio_message_columns=klio_message_columns
    )

    actual = mapper._map_row_element(row)

    assert actual == expected
